/*
 * Biting Blades 2 engine and library functions
 *
 */

#include "bladelib.h"
#include <stdio.h>

// Define extern consts from bladelib.h
const char location_name[7][10] = {"Head", "Right Arm", "Left Arm", "Chest",
                                   "Right Leg", "Left Leg", "Gut"};
const char location_abbr[7][3] = {"H", "RA", "LA", "C", "RL", "LL", "G"};

const int locations_adjacent[7][7] =
  {{1, 1, 1, 1, 0, 0, 0},   // head
   {1, 1, 0, 1, 1, 0, 0},   // right_arm
   {1, 0, 1, 1, 0, 1, 0},   // left_arm
   {1, 1, 1, 1, 0, 0, 1},   // chest
   {0, 1, 0, 0, 1, 1, 1},   // right_leg
   {0, 0, 1, 0, 1, 1, 1},   // left_leg
   {0, 0, 0, 1, 1, 1, 1}};  // gut

const char atk_spec_name[att_spec_count][10] = {"None"};
const char atk_spec_abbr[att_spec_count][3] = {"N"};

const char def_spec_name[def_spec_count][10] = {"Counter"};
const char def_spec_abbr[def_spec_count][3] = {"C"};

// Module-level variables
static player_info *head_player = 0;
static player_info *tail_player = 0;
static int current_tick = 0;

// Local function prototypes
static int inflict_damage(int hits, player_info *target);
static int generate_strength(int power);
static void insert_player(player_info *player, int impulse);
static void remove_player(player_info *player);
static void move_player(player_info *player, int new_impulse);
static void print_player_list();
static void remove_attack_card(player_info *player, attack_card *card);
static void remove_defense_card(player_info *player, defense_card *card);

extern void *malloc(size_t size);

/*
 * Inflicts speficied number of hits on *target and returns boolean
 * indication of whether *target is still alive
 */
static int inflict_damage(int hits, player_info *target) {
  int att;
  int def;
  int free;
  int extra;
  int i;

  /*
   * TO DO:  Decide whether Blades 2 will support extreme damamge effects
   *         (sever, crush, etc.) and, if so, implement them
   */

  // If no damage was inflicted, *target stays alive and there's nothing to do
  if (hits < 1) { return (1); }

  // If target took damage, make him wait another tick for each damage step
  // before his next impulse
  move_player(target, (target->next_impulse + hits));

  // Pull card allocations to internal storage for easy reference
  att = target->atk_avail;
  def = target->def_avail;
  free = target->free_avail;
  extra = target->null_steps;

  // Determine and apply each damage step's effect
  for (i = 1; i <= hits; i++) {
    if ((extra > 0) && (random() % 10)) {
      // If *target has any null_steps, each point of damage is 90% likely
      // to knock off one of them instead of a card
      extra--;
    } else if (att > ((def + 1) / 2)) {
      // Lose an attack card if the attack allocation is more than half the
      // defense allocation, rounded up
      att--;
    } else if (def >= ((free + 1) / 2)) {
      // Lose a defense card if the defense allocation is at least half of
      // the free allocation
      def--;
    } else {
      // Otherwise, convert a free card to a required defense
      free--;
      def++;
    }
  }

  // Copy updated data back into player
  target->atk_avail = att;
  target->def_avail = def;
  target->free_avail = free;
  target->null_steps = extra;

  // Return indication of whether character is still alive
  return ((att + def + free) > 0);
}

/*
 * Generate a random strength with an average value of power
 */
static int generate_strength(int power) {
  power++;
  
  return (random() % power) + (random() % power);
}

static void print_player_list() {
  player_info *current;

  printf("Current player order:\n");
  printf("head: %s(%d)\n", head_player ? head_player->name : "(null)", 
         head_player);
  current = head_player;
  while (current) {
    printf("%d - %s(%d) %s(%d) %s(%d)\n", current->next_impulse, current->name,
              current,
              current->prev_player ? current->prev_player->name : "(null)",
              current->prev_player,
              current->next_player ? current->next_player->name : "(null)",
              current->next_player);
    current = current->next_player;
  }
  printf("tail: %s(%d)\n\n", tail_player ? tail_player->name : "(null)",
         tail_player);
}

/*
 * Adds *player to the sequencing list on the tick for impulse
 */
static void insert_player(player_info *player, int impulse) {
  player_info *current;

  player->next_impulse = impulse;
  
  if (!head_player) {
    head_player         = player;
    tail_player         = player;
    player->prev_player = 0;
    player->next_player = 0;
  } else if (head_player->next_impulse > impulse) {
    player->prev_player      = 0;
    player->next_player      = head_player;
    head_player->prev_player = player;
    head_player              = player;
  } else if (tail_player->next_impulse <= impulse) {
    player->prev_player      = tail_player;
    player->next_player      = 0;
    tail_player->next_player = player;
    tail_player              = player;
  } else {
    current = head_player;
    while (current->next_player->next_impulse <= impulse) {
      current = current->next_player;
    }

    player->prev_player               = current;
    player->next_player               = current->next_player;
    current->next_player->prev_player = player;
    current->next_player              = player;
  }
}

/*
 * Removes *player from the sequencing list
 */
static void remove_player(player_info *player) {
  if (player->prev_player) {
    player->prev_player->next_player = player->next_player;
  } else {
    head_player = player->next_player;
  }

  if (player->next_player) {
    player->next_player->prev_player = player->prev_player;
  } else {
    tail_player = player->prev_player;
  }

  player->prev_player = 0;
  player->next_player = 0;

  if (!head_player) {
    exit(1);
  }
}

/*
 * Moves *player to new_impulse
 */
static void move_player(player_info *player, int new_impulse) {
  // There are more efficient ways to do this, but I'm going with the quick &
  // dirty solution for now.  Besides, it'll give insert_player and 
  // remove_player a good workout, which should help bring any bugs in them
  // to light.

  remove_player(player);
  insert_player(player, new_impulse);
}

static void remove_attack_card(player_info *player, attack_card *card) {
  attack_card *current = player->head_attack;

  if (player->head_attack == card) {
    player->head_attack = card->next_attack;
    card->next_attack = 0;
  } else {
    while (current) {
      if (current->next_attack == card) {
        current->next_attack = card->next_attack;
        card->next_attack = 0;
        break;
      }
      current = current->next_attack;
    }
  }
}

static void remove_defense_card(player_info *player, defense_card *card) {
  defense_card *current = player->head_defense;

  if (player->head_defense == card) {
    player->head_defense = card->next_defense;
    card->next_defense = 0;
  } else {
    while (current) {
      if (current->next_defense == card) {
        current->next_defense = card->next_defense;
        card->next_defense = 0;
        break;
      }
      current = current->next_defense;
    }
  }
}

/*
 * Verify that weapon is OK.  Currently just checks that attack and
 * defense target zone probabilities both add up to 100%.
 */
int validate_weapon(weapon_data weapon) {
  int atk_sum = 0;
  int def_sum = 0;
  int i;

  for (i = head; i <= gut; i++) {
    atk_sum += weapon.attack_chance[i];
    def_sum += weapon.defense_chance[i];
  }

  return ((atk_sum == 100) && (def_sum == 100));
}

/*
 * Used to reset/resynchronize impulse times at the start of a combat
 */
void reset_blades_sequencing() {
  current_tick = 0;
}

/*
 * Advances the tick counter by 1 and returns the new tick number
 */
int advance_tick_counter() {
  return (++current_tick);
}

/*
 * Returns the current time index
 */
int get_tick() {
  return (current_tick);
}

/*
 * Fills *the_list with pointers to up to max_players players and returns the
 * number of players inserted
 */
int list_players(player_info **the_list, int max_players) {
  player_info *current = head_player;
  int i = 0;

  while ((current) && (i < max_players)) {
    (*the_list) = current;

    current = current->next_player;
    (the_list)++;
    i++;
  }

  return(i);
}

/*
 * Returns the number of active players
 */
int player_count() {
  player_info *current;
  int i = 0;

  if (head_player) {
    current = head_player;
    i = 1;
    while (current->next_player) {
      i++;
      current = current->next_player;
    }
  }

  return(i);
}

/*
 * Inserts player into the impulse sequence so that he'll be able to act
 */
void add_player(player_info *player) {
  insert_player(player, current_tick + (random() % 10) + 1);
}

/*
 * Removes any of player's defenses which may have expired
 */
void clear_defense(player_info *player) {
  defense_card *curr_def;
  defense_card *prev_card; 

  // When a defense is played, it gets moved to the head of the hand, so
  // once we find an unplayed card, we're done
  curr_def = player->head_defense;
  while ((curr_def) && (curr_def->expires)) {
    prev_card = curr_def;
    curr_def = curr_def->next_defense;
    if (prev_card->expires < current_tick) {
      remove_defense_card(player, prev_card);
    }
  }
}

/*
 * Returns a pointer to the player whose impulse is currently active or null
 * if no further players may act this tick
 */
player_info *current_player() {
  defense_card *curr_def;
  defense_card *prev_card; 

  if ((!head_player) || (head_player->next_impulse > current_tick)) {
    return(0);
  }

  // When your impulse comes up, all defenses expire

  // When a defense is played, it gets moved to the head of the hand, so
  // we just have to go through defense cards until we find an unplayed
  // one and trash everything up to (but not including) that point
  curr_def = head_player->head_defense;
  while ((curr_def) && (curr_def->expires)) {
    prev_card = curr_def;
    curr_def = curr_def->next_defense;
    free(prev_card);
  }
  head_player->head_defense = curr_def;
 
  return(head_player);
}

/*
 * Advance tick counter until someone gets an impulse and return a pointer
 * to that player
 */
player_info *next_player() {
  if (head_player) {
    current_tick = head_player->next_impulse;
  }

  return(current_player());
}

/*
 * Make player wait the specified number of ticks before acting
 */
void delay_player(player_info *player, int ticks) {
  if ((current_tick + ticks) > player->next_impulse) {
    move_player(player, current_tick + ticks);
  }
}
  
/* 
 * Generate an attack card for weapon and place it in *player's hand
 */
void generate_attack_card(player_info *player, const weapon_data *weapon) {
  attack_card *card;
  body_location loc_index;
  int i;

  card = malloc(sizeof(attack_card));
  card->next_attack = player->head_attack;
  player->head_attack = card;
  /*
   * Should zero out the new card!
   */

  // Set card's weapon
  card->weapon = weapon;
  
  // Figure out where the attack is targeted
  i = (random() % 100);
  loc_index = head - 1;
  while (i > 0) {
    i -= weapon->attack_chance[++loc_index];
  }
  card->location = loc_index;

  // Determine attack strength
  card->strength = generate_strength(weapon->atk_value);

  // Run through the list of attack specials and see which are present
  for (i = 0; i < att_spec_count; i++) {
    card->special[i] = ((random() % 100) < weapon->atk_spec_chance[i]);
  }
}

/*
 * Generate a defense card for weapon and place it in *player's hand
 */
void generate_defense_card(player_info *player, const weapon_data *weapon) {
  defense_card *card;
  body_location loc_index;
  body_location center;
  int i;

  card = malloc(sizeof(defense_card));
  card->next_defense = player->head_defense;
  player->head_defense = card;
  /*
   * Should zero out the new card!
   */

  // Set card's weapon
  card->weapon = weapon;
  card->duration = (weapon->speed / 2);
  card->expires = 0;
  
  // Figure out where the defense is centered
  i = (random() % 100);
  center = head - 1;
  while (i > 0) {
    i -= weapon->defense_chance[++center];
  }

  // Determine defense strengths
  for (loc_index = head; loc_index <= gut; loc_index++) {
    if (loc_index == center) {
      card->strength[loc_index] = generate_strength(weapon->pri_def_value);
    } else if (locations_adjacent[loc_index][center]) {
      card->strength[loc_index] = generate_strength(weapon->adj_def_value);
    } else {
      card->strength[loc_index] = 0;
    }
  }

  // Run through the list of defense specials and see which are present
  for (i = 0; i < def_spec_count; i++) {
    card->special[i] = ((random() % 100) < weapon->def_spec_chance[i]);
  }
}

/*
 * Determine and apply the effects of attack on *target, who is using defense
 */
int resolve_attack(player_info *attacker, attack_card  *attack,
                   player_info *target,   defense_card *defense) {
  attack_card *curr_att;
  defense_card *curr_def;

  int damage = attack->strength;
  int def_speed = 0;
  int next_def_impulse = current_tick + 1;

  // Mark the defense as being in effect and move to head of hand
  if (defense) {
    defense->expires = current_tick + defense->duration;
    remove_defense_card(target, defense);
    defense->next_defense = target->head_defense;
    target->head_defense = defense;

    def_speed = defense->special[counterattack] ?
                                 0 : (defense->weapon->speed / 2 + 1);
    next_def_impulse = current_tick + def_speed;
  }

  // Reduce attack strength for all active defenses
  curr_def = target->head_defense;
  while ((curr_def) && (curr_def->expires)) {
    damage -= curr_def->strength[attack->location];
    curr_def = curr_def->next_defense;
  }

  // Anyone who is attacked is considered engaged until their next impulse
  if (def_speed >= target->engaged) {
    target->engaged = def_speed;
  }

  // Attacking uses up the full number of ticks from the weapon's speed
  move_player(attacker, current_tick + attack->weapon->speed);
  remove_attack_card(attacker, attack);

  // Defending moves your next impulse out to now plus half your weapon's
  // speed in ticks, unless it's already further out than that
  if (next_def_impulse > target->next_impulse) {
    move_player(target, next_def_impulse);
  }

  return(inflict_damage(damage, target));
}
