/*
 * bladelib.h
 *
 * Biting Blades 2 engine and library functions
 */

#ifndef __bladelib__
#define __bladelib__

// List of hit locations in various formats
enum body_location_enum {head, right_arm, left_arm, chest, right_leg, 
                         left_leg, gut};
typedef enum body_location_enum body_location;
extern const char location_name[7][10];
extern const char location_abbr[7][3];

// Map of which locations are adjacent to each other
extern const int locations_adjacent[7][7];

// List of attack specials in various formats
enum {att_spec_count = 1};
enum attack_special_enum {null_attack_special};
enum {min_att_spec = null_attack_special, max_att_spec = null_attack_special};
typedef enum attack_special_enum attack_special;
extern const char atk_spec_name[att_spec_count][10];
extern const char atk_spec_abbr[att_spec_count][3];

// List of defense specials in various formats
enum {def_spec_count = 1};
enum defense_special_enum {counterattack};
enum {min_def_spec = counterattack, max_def_spec = counterattack};
typedef enum defense_special_enum defense_special;
extern const char def_spec_name[def_spec_count][10];
extern const char def_spec_abbr[def_spec_count][3];

// Structure defining a weapon's characteristics
struct weapon_data_struct {
  char name[10];                // Name of the weapon

  int speed;                    // Number of ticks consumed by attacking

  unsigned char atk_value;      // Average attack strength
  unsigned char pri_def_value;  // Average primary defense
  unsigned char adj_def_value;  // Average adjacent defense

  // % chance of attack/defense focusing on each location
  unsigned char attack_chance[7];
  unsigned char defense_chance[7];

  // % chance of each special attribute being present
  // *** Later will need to add info on additional data for each ***
  // *** special (e.g., amount of extra strength for frenzy,     ***
  // *** strength reduction for counter or feint, etc.)          ***
  unsigned char atk_spec_chance[att_spec_count];
  unsigned char def_spec_chance[def_spec_count];
};
typedef struct weapon_data_struct weapon_data;

// Structure describing a single attack card
struct attack_card_struct {
  const weapon_data *weapon;            // Pointer to the weapon
  
  body_location location;         // Location targeted by attack
  unsigned char strength;         // Strength of attack
  char special[att_spec_count];   // Booleans for presence of specials

  struct attack_card_struct *next_attack;
};
typedef struct attack_card_struct attack_card;

// Structure describing a single defense card
struct defense_card_struct {
  const weapon_data *weapon;            // Pointer to the weapon

  int duration;               // How long does this card protect for?
  int expires;                // On what tick does this card stop working?

  unsigned char strength[7];      // Strength of defense by location
  char special[def_spec_count];   // Booleans for presence of specials

  struct defense_card_struct *next_defense;
};
typedef struct defense_card_struct defense_card;

// Structure describing a character in the combat
struct player_info_struct {
  char name[20];    // Character's name

  int next_impulse; // When do I get to do something?
  int engaged;      // Is character currently engaged in melee?

  int atk_avail;    // Attack picks available
  int def_avail;    // Defense picks available
  int free_avail;   // Free picks available
  int null_steps;   // Null picks available

  int atk_cards;    // Number of attack cards held
  int def_cards;    // Number of defense cards held

  attack_card  *head_attack;      // Data on attack cards in hand
  defense_card *head_defense;     // Data on defense cards in hand

  struct player_info_struct *prev_player;
  struct player_info_struct *next_player;
};
typedef struct player_info_struct player_info;

/*
 * Verify that a weapon data struct is OK.
 */
extern int validate_weapon(weapon_data weapon);

/*
 * Used to reset/resynchronize impulse times at the start of a combat
 */
extern void reset_blades_sequencing();

/* 
 * Returns current tick
 */
extern int get_tick();

/*
 * Advances the tick counter by 1 and returns the new tick number
 */
extern int advance_tick_counter();

/*
 * Inserts player into the impulse sequence so that he'll be able to act
 */
extern void add_player(player_info *player);

/*
 * Fills *the_list with pointers to up to max_players players and returns the
 * number of players inserted
 */
extern int list_players(player_info **the_list, int max_players);

/*
 * Removes any of player's defenses which may have expired
 */
extern void clear_defense(player_info *player);

/*
 * Returns a pointer to the player whose impulse is currently active or null
 * if no further players may act this tick
 */
extern player_info *current_player();

/*
 * Advance tick counter until someone gets an impulse and return a pointer
 * to that player
 */
extern player_info *next_player();

/*
 * Make player wait the specified number of ticks before acting
 */
extern void delay_player(player_info *player, int ticks);

/*
 * Returns the number of active players
 */
extern int player_count();

/* 
 * Creates an attack card for weapon and adds it to *player's hand
 */
extern void generate_attack_card(player_info *player,
                                 const weapon_data *weapon);

/* 
 * Creates a defense card for weapon and adds it to *player's hand
 */
extern void generate_defense_card(player_info *player,
                                  const weapon_data *weapon);

/* 
 * Determines and applies the effect of *attacker's *attack on *target, who
 * is using *defense
 */
extern int resolve_attack(player_info *attacker, attack_card  *attack,
                          player_info *target,   defense_card *defense);

#endif
